#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Objects that utilize command-line inputs to run a program using Nunavut.
"""
import argparse
import logging
import pathlib
import sys
import textwrap
import typing

import nunavut
import nunavut.jinja
import nunavut.lang
import pydsdl
from nunavut.generators import AbstractGenerator, create_generators
from nunavut._utilities import YesNoDefault


class ArgparseRunner:
    """
    Runner that uses Python argparse arguments to define a run.

    :param argparse.Namespace args: The commandline arguments.
    :param typing.Optional[typing.Union[str, typing.List[str]]] extra_includes: A list of paths to additional DSDL
        root folders.
    """

    def __init__(self, args: argparse.Namespace, extra_includes: typing.Optional[typing.Union[str, typing.List[str]]]):
        self._args = args
        self._generators = None  # type: typing.Optional[typing.Tuple[AbstractGenerator, AbstractGenerator]]
        self._root_namespace = None  # type: typing.Optional[nunavut.Namespace]

        if extra_includes is None:
            extra_includes = []
        elif not isinstance(extra_includes, list):
            extra_includes = [extra_includes]

        self._extra_includes = extra_includes

    @property
    def extra_includes(self) -> typing.List[str]:
        return self._extra_includes

    @property
    def generator(self) -> AbstractGenerator:
        if self._generators is None:
            raise RuntimeError("generator property accessed before setup")
        return self._generators[0]

    @property
    def support_generator(self) -> AbstractGenerator:
        if self._generators is None:
            raise RuntimeError("support_generator property accessed before setup")
        return self._generators[1]

    @property
    def root_namespace(self) -> nunavut.Namespace:
        if self._root_namespace is None:
            raise RuntimeError("root_namespace property accessed before setup")
        return self._root_namespace

    def setup(self) -> None:
        """
        Required to prepare this object to run (run method will raise exceptions if called before this method).
        While this may seem a bit clunky it helps isolate errors to two distinct stages; setup and run.

        Setup never generates anything. It only parses the inputs and creates the generator arguments.
        """

        #
        # nunavut : parse inputs
        #
        language_context = self._create_language_context()

        if self._args.generate_support != "only":
            type_map = pydsdl.read_namespace(
                self._args.root_namespace,
                self._extra_includes,
                allow_unregulated_fixed_port_id=self._args.allow_unregulated_fixed_port_id,
            )
        else:
            type_map = []

        self._root_namespace = nunavut.build_namespace_tree(
            type_map, self._args.root_namespace, self._args.outdir, language_context
        )

        #
        # nunavut : create generators
        #

        generator_args = {
            "generate_namespace_types": (
                YesNoDefault.YES if self._args.generate_namespace_types else YesNoDefault.DEFAULT
            ),
            "templates_dir": (pathlib.Path(self._args.templates) if self._args.templates is not None else None),
            "trim_blocks": self._args.trim_blocks,
            "lstrip_blocks": self._args.lstrip_blocks,
            "post_processors": self._build_post_processor_list_from_args(),
        }

        self._generators = create_generators(self._root_namespace, **generator_args)

    def run(self) -> None:
        """
        Perform actions defined by the arguments this object was created with. This may generate outputs where
        the arguments have requested this action.

        .. warning::
            :meth:`setup` must be called before calling this method.

        """

        if self._args.list_outputs:
            self._list_outputs_only()

        elif self._args.list_inputs:
            self._list_inputs_only()

        else:
            self._generate()

    # +---------------------------------------------------------------------------------------------------------------+
    # | PRIVATE
    # +---------------------------------------------------------------------------------------------------------------+

    def _should_generate_support(self) -> bool:
        if self._args.generate_support == "as-needed":
            return self._args.omit_serialization_support is None or not self._args.omit_serialization_support
        else:
            return bool(self._args.generate_support == "always" or self._args.generate_support == "only")

    def _build_ext_program_postprocessor(self, program: str) -> nunavut.postprocessors.FilePostProcessor:
        subprocess_args = [program]
        if hasattr(self._args, "pp_run_program_arg") and self._args.pp_run_program_arg is not None:
            for program_arg in self._args.pp_run_program_arg:
                subprocess_args.append(program_arg)
        return nunavut.postprocessors.ExternalProgramEditInPlace(subprocess_args)

    def _build_post_processor_list_from_args(self) -> typing.List[nunavut.postprocessors.PostProcessor]:
        """
        Return a list of post processors setup based on the provided command-line arguments. This
        list may be empty but the function will not return None.
        """
        post_processors = []  # type: typing.List[nunavut.postprocessors.PostProcessor]
        if self._args.pp_trim_trailing_whitespace:
            post_processors.append(nunavut.postprocessors.TrimTrailingWhitespace())
        if hasattr(self._args, "pp_max_emptylines") and self._args.pp_max_emptylines is not None:
            post_processors.append(nunavut.postprocessors.LimitEmptyLines(self._args.pp_max_emptylines))
        if hasattr(self._args, "pp_run_program") and self._args.pp_run_program is not None:
            post_processors.append(self._build_ext_program_postprocessor(self._args.pp_run_program))

        post_processors.append(nunavut.postprocessors.SetFileMode(self._args.file_mode))

        return post_processors

    def _create_language_context(self) -> nunavut.lang.LanguageContext:
        language_options = dict()
        if self._args.target_endianness is not None:
            language_options["target_endianness"] = self._args.target_endianness
        language_options["omit_float_serialization_support"] = self._args.omit_float_serialization_support
        language_options["enable_serialization_asserts"] = self._args.enable_serialization_asserts
        language_options["enable_override_variable_array_capacity"] = self._args.enable_override_variable_array_capacity
        if self._args.language_standard is not None:
            language_options["std"] = self._args.language_standard

        language_context = nunavut.lang.LanguageContext(
            self._args.target_language,
            self._args.output_extension,
            self._args.namespace_output_stem,
            omit_serialization_support_for_target=self._args.omit_serialization_support,
            language_options=language_options,
            include_experimental_languages=self._args.experimental_languages,
        )

        #
        # nunavut: inferred target language from extension
        #
        if self._args.output_extension is not None and language_context.get_target_language() is None:

            inferred_target_language_name = None  # type: typing.Optional[str]
            for name, lang in language_context.get_supported_languages().items():
                extension = lang.get_config_value("extension", None)
                if extension is not None and extension == self._args.output_extension:
                    inferred_target_language_name = name
                    break

            if inferred_target_language_name is not None:
                logging.info(
                    'Inferring target language %s based on extension "%s".',
                    inferred_target_language_name,
                    self._args.output_extension,
                )
                language_context = nunavut.lang.LanguageContext(
                    inferred_target_language_name,
                    self._args.output_extension,
                    self._args.namespace_output_stem,
                    omit_serialization_support_for_target=self._args.omit_serialization_support,
                    language_options=language_options,
                )
            elif self._args.templates is None:
                logging.warn(
                    textwrap.dedent(
                        """
                    ***********************************************************************
                        No target language was given, none could be inferred from the output extension (-e) argument
                        "%s", and no user templates were specified. You will fail to find templates if you have provided
                        any DSDL types to generate.
                    ***********************************************************************
                    """
                    ).lstrip(),
                    self._args.output_extension,
                )
        return language_context

    # +---------------------------------------------------------------------------------------------------------------+
    # | PRIVATE :: RUN METHODS
    # +---------------------------------------------------------------------------------------------------------------+
    def _stdout_lister(
        self, things_to_list: typing.Iterable[typing.Any], to_string: typing.Callable[[typing.Any], str]
    ) -> None:
        for thing in things_to_list:
            sys.stdout.write(to_string(thing))
            sys.stdout.write(";")

    def _list_outputs_only(self) -> None:
        if self._args.generate_support != "only":
            self._stdout_lister(self.generator.generate_all(is_dryrun=True), lambda p: str(p))

        if self._should_generate_support():
            self._stdout_lister(self.support_generator.generate_all(is_dryrun=True), lambda p: str(p))

    def _list_inputs_only(self) -> None:
        if self._args.generate_support != "only":
            self._stdout_lister(self.generator.get_templates(), lambda p: str(p.resolve()))

        if self._should_generate_support():
            self._stdout_lister(self.support_generator.get_templates(), lambda p: str(p.resolve()))

        if self._args.generate_support != "only":
            if self.generator.generate_namespace_types:
                self._stdout_lister(
                    [x for x, _ in self.root_namespace.get_all_types()], lambda p: p.source_file_path.as_posix()
                )
            else:
                self._stdout_lister(
                    [x for x, _ in self.root_namespace.get_all_datatypes()], lambda p: p.source_file_path.as_posix()
                )

    def _generate(self) -> None:
        if self._should_generate_support():
            self.support_generator.generate_all(
                is_dryrun=self._args.dry_run, allow_overwrite=not self._args.no_overwrite
            )

        if self._args.generate_support != "only":
            self.generator.generate_all(is_dryrun=self._args.dry_run, allow_overwrite=not self._args.no_overwrite)

        # TODO: move this somewhere html-specific.
        if self._args.target_language == "html" and len(self.extra_includes) > 0:
            logging.warning(
                "Other lookup namespaces are linked in these generated docs. "
                "If you do not generate docs for these other namespaces as well, "
                "links to external data types could be broken (expansion will still work)."
            )



#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
import datetime
import inspect
import logging
import types
import typing

import nunavut.lang

from ..templates import LanguageEnvironment
from .extensions import JinjaAssert, UseQuery
from .jinja2 import BaseLoader, Environment, StrictUndefined, select_autoescape
from .jinja2.ext import Extension
from .jinja2.ext import do as jinja_do
from .jinja2.ext import loopcontrols as loopcontrols
from .jinja2.filters import FILTERS as JINJA2_FILTERS

logger = logging.getLogger(__name__)

# +---------------------------------------------------------------------------+
# | JINJA : LanguageTemplateNamespace
# +---------------------------------------------------------------------------+


class LanguageTemplateNamespace:
    """
    Generic namespace object used to create reserved namespaces in the global environment.

    .. invisible-code-block: python

        from nunavut.jinja.environment import LanguageTemplateNamespace

    .. code-block:: python

        ns = LanguageTemplateNamespace()

        # any property can be set at any time.
        ns.foo = 'foo'
        assert ns.foo == 'foo'

        # repr of the ns enables cloning using exec
        exec('ns2={}'.format(repr(ns)))
        assert ns2.foo == 'foo'

        # clones will be equal
        assert ns2 == ns

        # but not the same object
        assert ns2 is not ns

    In addition to the namespace behavior this object exposed some dictionary-like methods:

    .. code-block:: python

        ns = LanguageTemplateNamespace()
        ns.update({'foo':'bar'})

        assert ns.foo == 'bar'

    .. invisible-code-block: python

        ns = LanguageTemplateNamespace(one='one', two='two')
        assert 'one' == ns.one
        assert 'two' == ns.two

    """

    def __init__(self, **kwargs: typing.Any):
        for name in kwargs:
            setattr(self, name, kwargs[name])

    def __repr__(self) -> str:
        type_name = type(self).__name__
        arg_strings = []
        star_args = {}
        for name, value in self._get_kwargs():
            if name.isidentifier():
                arg_strings.append("%s=%r" % (name, value))
            else:
                star_args[name] = value
        if star_args:
            arg_strings.append("**%s" % repr(star_args))
        return "%s(%s)" % (type_name, ", ".join(arg_strings))

    def _get_kwargs(self) -> typing.List[typing.Any]:
        return list(self.__dict__.items())

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, LanguageTemplateNamespace):
            return NotImplemented
        return vars(self) == vars(other)

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def update(self, update_from: typing.Mapping[str, typing.Any]) -> None:
        for key, value in update_from.items():
            setattr(self, key, value)

    def items(self) -> typing.ItemsView[str, typing.Any]:
        return self.__dict__.items()

    def values(self) -> typing.ValuesView[typing.Any]:
        return self.__dict__.values()


# +---------------------------------------------------------------------------+
# | JINJA : CodeGenEnvironment
# +---------------------------------------------------------------------------+


class CodeGenEnvironment(Environment):
    """
    Jinja Environment optimized for compile-time generation of source code
    (i.e. as opposed to dynamically generating webpages).

    .. invisible-code-block: python

        from nunavut.lang import LanguageContext, Language
        from nunavut.jinja import CodeGenEnvironment
        from nunavut.jinja.jinja2 import DictLoader

    .. code-block:: python

        template = 'Hello World'

        e = CodeGenEnvironment(loader=DictLoader({'test': template}))
        assert 'Hello World' ==  e.get_template('test').render()

    .. warning::
        The :attr:`RESERVED_GLOBAL_NAMESPACES` and :attr:`RESERVED_GLOBAL_NAMES` collections
        contain names in the global namespace reserved by this environment. Attempting to override one
        of these reserved names will cause the constructor to raise an error.

    .. code-block:: python

        try:
            CodeGenEnvironment(loader=DictLoader({'test': template}), additional_globals={'ln': 'bad_ln'})
            assert False
        except RuntimeError:
            pass

    Other safe-guards include checks that Jinja built-ins aren't accidentally overridden...

    .. code-block:: python

        try:
            CodeGenEnvironment(loader=DictLoader({'test': template}),
                                                 additional_filters={'indent': lambda x: x})
            assert False
        except RuntimeError:
            pass

        # You can allow overwrite of built-ins using the ``allow_filter_test_or_use_query_overwrite``
        # argument.
        e = CodeGenEnvironment(loader=DictLoader({'test': template}),
                                                 additional_filters={'indent': lambda x: x},
                                                 allow_filter_test_or_use_query_overwrite=True)
        assert 'foo' == e.filters['indent']('foo')

    ...or that user-defined filters or redefined.

    .. code-block:: python

        class MyFilters:

            @staticmethod
            def filter_misnamed(name: str) -> str:
                return name

        e = CodeGenEnvironment(loader=DictLoader({'test': template}),
                               additional_filters={'filter_misnamed': lambda x: x})

        try:
            e.add_conventional_methods_to_environment(MyFilters())
            assert False
        except RuntimeError:
            pass

    .. note:: Maintainer's Note
        This class should remain DSDL agnostic. It is, theoretically, applicable using Jinja with any compiler front-end
        input although, in practice, it will only ever be used with pydsdl AST.
        Pydsdl-specific logic should live in the CodeGenerator (:class:`nunavut.jinja.DSDLCodeGenerator`).

    """

    RESERVED_GLOBAL_NAMESPACES = {"ln", "options", "uses_queries", "nunavut"}

    RESERVED_GLOBAL_NAMES = {"now_utc"}

    NUNAVUT_NAMESPACE_PREFIX = "nunavut.lang."

    def __init__(
        self,
        loader: BaseLoader,
        lctx: typing.Optional[nunavut.lang.LanguageContext] = None,
        trim_blocks: bool = False,
        lstrip_blocks: bool = False,
        additional_filters: typing.Optional[typing.Dict[str, typing.Callable]] = None,
        additional_tests: typing.Optional[typing.Dict[str, typing.Callable]] = None,
        additional_globals: typing.Optional[typing.Dict[str, typing.Any]] = None,
        extensions: typing.List[Extension] = [jinja_do, loopcontrols, JinjaAssert, UseQuery],
        allow_filter_test_or_use_query_overwrite: bool = False,
    ):
        super().__init__(
            loader=loader,  # nosec
            extensions=extensions,
            autoescape=select_autoescape(
                enabled_extensions=("htm", "html", "xml", "json"), default_for_string=False, default=False
            ),
            undefined=StrictUndefined,
            keep_trailing_newline=True,
            lstrip_blocks=lstrip_blocks,
            trim_blocks=trim_blocks,
            auto_reload=False,
            cache_size=400,
        )
        if additional_globals is not None:
            for global_name, global_value in additional_globals.items():
                if global_name in self.RESERVED_GLOBAL_NAMESPACES or global_name in self.RESERVED_GLOBAL_NAMES:
                    raise RuntimeError('Additional global "{}" uses a reserved global name'.format(global_name))
                self.globals[global_name] = global_value

        self._allow_replacements = allow_filter_test_or_use_query_overwrite

        for global_namespace in self.RESERVED_GLOBAL_NAMESPACES:
            self.globals[global_namespace] = LanguageTemplateNamespace()

        self.globals["now_utc"] = datetime.datetime(datetime.MINYEAR, 1, 1)
        self._target_language = None  # type: typing.Optional[nunavut.lang.Language]

        # --------------------------------------------------
        # After this point we do that most heinous act so common in dynamic languages;
        # we expose the state of this partially constructed object so we can complete
        # configuring it.

        if lctx is not None:
            self._update_language_support(lctx)

            supported_languages = (
                lctx.get_supported_languages().values()
            )  # type: typing.Optional[typing.ValuesView[nunavut.lang.Language]]
        else:
            supported_languages = None

        self._update_nunavut_globals(lctx)

        self.add_conventional_methods_to_environment(self)

        if additional_filters is not None:
            self._add_each_to_environment(
                additional_filters.items(), self.filters, supported_languages=supported_languages
            )
        if additional_tests is not None:
            self._add_each_to_environment(additional_tests.items(), self.tests, supported_languages=supported_languages)

    def add_conventional_methods_to_environment(self, obj: typing.Any) -> None:
        for name, method in inspect.getmembers(obj, inspect.isroutine):
            try:
                self._add_conventional_method_to_environment(method, name, supported_languages=self.supported_languages)
            except TypeError:
                pass

    @property
    def supported_languages(self) -> typing.ValuesView[nunavut.lang.Language]:
        ln_globals = self.globals["ln"]  # type: LanguageTemplateNamespace
        return ln_globals.values()

    @property
    def nunavut_global(self) -> LanguageTemplateNamespace:
        return typing.cast(LanguageTemplateNamespace, self.globals["nunavut"])

    @property
    def target_language_uses_queries(self) -> LanguageTemplateNamespace:
        return typing.cast(LanguageTemplateNamespace, self.globals["uses_queries"])

    @property
    def language_options(self) -> LanguageTemplateNamespace:
        return typing.cast(LanguageTemplateNamespace, self.globals["options"])

    @property
    def language_support(self) -> LanguageTemplateNamespace:
        return typing.cast(LanguageTemplateNamespace, self.globals["ln"])

    @property
    def target_language(self) -> typing.Optional[nunavut.lang.Language]:
        return self._target_language

    @property
    def now_utc(self) -> datetime.datetime:
        return typing.cast(datetime.datetime, self.globals["now_utc"])

    @now_utc.setter
    def now_utc(self, utc_time: datetime.datetime) -> None:
        self.globals["now_utc"] = utc_time

    def add_test(self, test_name: str, test_callable: typing.Callable) -> None:
        self._add_to_environment(test_name, test_callable, self.tests)

    # +----------------------------------------------------------------------------------------------------------------+
    # | Private
    # +----------------------------------------------------------------------------------------------------------------+
    def _resolve_collection(
        self,
        conventional_method_prefix: typing.Optional[str],
        method_name: str,
        collection_maybe: typing.Optional[typing.Union[LanguageTemplateNamespace, typing.Dict[str, typing.Any]]],
    ) -> typing.Union[LanguageTemplateNamespace, typing.Dict[str, typing.Any]]:
        if collection_maybe is not None:
            return collection_maybe

        if LanguageEnvironment.is_test_name(conventional_method_prefix):
            return typing.cast(typing.Dict[str, typing.Any], self.tests)
        elif LanguageEnvironment.is_filter_name(conventional_method_prefix):
            return typing.cast(typing.Dict[str, typing.Any], self.filters)
        elif LanguageEnvironment.is_uses_query_name(conventional_method_prefix):
            uses_queries = self.globals["uses_queries"]
            return typing.cast(LanguageTemplateNamespace, uses_queries)
        else:
            raise TypeError(
                "Tried to add an item {} to the template environment but we don't know what the item is.".format(
                    method_name
                )
            )

    def _add_to_environment(
        self,
        item_name: str,
        item: typing.Any,
        collection: typing.Union[LanguageTemplateNamespace, typing.Dict[str, typing.Any]],
    ) -> None:
        if item_name in collection:
            if not self._allow_replacements:
                raise RuntimeError("{} was already defined.".format(item_name))
            elif item_name in JINJA2_FILTERS:
                logger.info("Replacing Jinja built-in {}".format(item_name))
            else:
                logger.info('Replacing "{}" which was already defined for this environment.'.format(item_name))
        else:
            logger.debug("Adding {} to environment".format(item_name))
        if isinstance(collection, LanguageTemplateNamespace):
            setattr(collection, item_name, item)
        else:
            collection[item_name] = item

    def _add_conventional_method_to_environment(
        self,
        method: typing.Callable[..., bool],
        method_name: str,
        collection_maybe: typing.Optional[typing.Union[LanguageTemplateNamespace, typing.Dict[str, typing.Any]]] = None,
        supported_languages: typing.Optional[typing.ValuesView[nunavut.lang.Language]] = None,
        method_language: typing.Optional[nunavut.lang.Language] = None,
        is_target: bool = False,
    ) -> None:
        """

        :param str callable_name: The name of the callable to use in a template.
        :param typing.Callable[..., bool] callable: The named callable.
        :param typing.Optional[str] callable_namespace: If provided the namespace to prefix to the callable name.
        :return: tuple of name and the callable which might be prepared as a partial function based on decorators.
        :raises: RuntimeWarning if the callable requested resources that were not available in this environment.

            .. invisible-code-block: python

                from nunavut.jinja import CodeGenEnvironment
                from nunavut.jinja.jinja2 import DictLoader
                from nunavut.templates import template_language_test
                from unittest.mock import MagicMock

                lctx = MagicMock(spec=LanguageContext)
                poop_lang = MagicMock(spec=Language)
                poop_lang.name = 'poop'
                poop_lang.get_templates_package_name = MagicMock(return_value='nunavut.lang.poop')
                lctx.get_target_language = MagicMock(return_value=None)
                lctx.get_supported_languages = MagicMock(return_value = {'poop': poop_lang})

                @template_language_test('nunavut.lang.poop')
                def test_test(language):
                    return True

                e = CodeGenEnvironment(
                    loader=DictLoader({'test': 'hello world'}),
                    additional_tests={'foo': test_test},
                    lctx=lctx
                )
                assert test_test == e.tests['foo'].func
                assert e.tests['foo']()
        """

        result = LanguageEnvironment.handle_conventional_methods(method, method_name, supported_languages)
        collection = self._resolve_collection(result[0], method_name, collection_maybe)

        if method_language is not None:
            self._add_to_environment("ln.{}.{}".format(method_language.name, result[1]), result[2], collection)
        else:
            self._add_to_environment(result[1], result[2], collection)
        if is_target:
            self._add_to_environment(result[1], result[2], collection)

    def _add_each_to_environment(
        self,
        items: typing.AbstractSet[typing.Tuple[str, typing.Callable]],
        collection: typing.Optional[
            typing.Union[
                LanguageTemplateNamespace,
                typing.Dict[str, typing.Any],
            ]
        ] = None,
        supported_languages: typing.Optional[typing.ValuesView[nunavut.lang.Language]] = None,
        language: typing.Optional[nunavut.lang.Language] = None,
        is_target: bool = False,
    ) -> None:
        for method_name, method in items:
            self._add_conventional_method_to_environment(
                method, method_name, collection, supported_languages, language, is_target
            )

    @classmethod
    def _create_platform_version(cls) -> typing.Dict[str, typing.Any]:
        import platform
        import sys

        platform_version = {}  # type: typing.Dict[str, typing.Any]

        platform_version["python_implementation"] = platform.python_implementation()
        platform_version["python_version"] = platform.python_version()
        platform_version["python_release_level"] = sys.version_info[3]
        platform_version["python_build"] = platform.python_build()
        platform_version["python_compiler"] = platform.python_compiler()
        platform_version["python_revision"] = platform.python_revision()

        try:
            platform_version["python_xoptions"] = sys._xoptions
        except AttributeError:  # pragma: no cover
            platform_version["python_xoptions"] = {}

        platform_version["runtime_platform"] = platform.platform()

        return platform_version

    def _add_support_from_language_module_to_environment(
        self,
        lctx: nunavut.lang.LanguageContext,
        language: nunavut.lang.Language,
        ln_module: "types.ModuleType",
        is_target: bool = False,
    ) -> None:
        supported_languages = lctx.get_supported_languages()
        ln_env = LanguageEnvironment.find_all_conventional_methods_in_language_module(
            language, supported_languages.values(), ln_module
        )
        self._add_each_to_environment(
            ln_env.filters.items(), self.filters, supported_languages.values(), language=language, is_target=is_target
        )
        self._add_each_to_environment(
            ln_env.tests.items(), self.tests, supported_languages.values(), language=language, is_target=is_target
        )
        if is_target:
            self._target_language = language
            self._add_each_to_environment(
                ln_env.uses_queries.items(),
                self.globals["uses_queries"],
                supported_languages.values(),
                language=language,
                is_target=is_target,
            )

    def _update_language_support(self, lctx: nunavut.lang.LanguageContext) -> None:

        supported_languages = lctx.get_supported_languages()
        target_language = lctx.get_target_language()
        ln_globals = self.globals["ln"]
        if target_language is not None:
            self.globals.update(target_language.get_globals())
            globals_options_ns = self.globals["options"]
            globals_options_ns.update(target_language.get_options())
        for supported_language in supported_languages.values():
            if supported_language.name not in ln_globals:
                setattr(
                    ln_globals, supported_language.name, LanguageTemplateNamespace(options=LanguageTemplateNamespace())
                )
            ln_globals_ns = getattr(ln_globals, supported_language.name)
            ln_globals_ns.update(supported_language.get_globals())
            ln_globals_options_ns = getattr(ln_globals_ns, "options")
            ln_globals_options_ns.update(supported_language.get_options())

        # then load everything into the environment from this list.
        # note that we don't unload anything here so this method is not idempotent
        for supported_language in supported_languages.values():
            is_target = target_language is not None and supported_language == target_language
            try:
                self._add_support_from_language_module_to_environment(
                    lctx,
                    supported_language,
                    nunavut.lang.LanguageLoader.load_language_module(supported_language.name),
                    is_target,
                )
            except ModuleNotFoundError:
                pass

    def _update_nunavut_globals(self, lctx: typing.Optional[nunavut.lang.LanguageContext] = None) -> None:

        # Helper global so we don't have to futz around with the "omit_serialization_support"
        # logic in the templates. The omit_serialization_support property of the Language
        # object is read-only so this boolean will remain consistent for the Environment.
        target_language = None if lctx is None else lctx.get_target_language()
        if target_language is not None:
            omit_serialization_support = target_language.omit_serialization_support
            support_namespace, support_version, _ = target_language.get_support_module()
        else:
            logger.debug("There is no target language so we cannot generate serialization support")
            omit_serialization_support = True
            support_namespace = ""
            support_version = (0, 0, 0)

        nunavut_namespace = self.nunavut_global

        setattr(
            nunavut_namespace,
            "support",
            {"omit": omit_serialization_support, "namespace": support_namespace, "version": support_version},
        )

        if "version" not in nunavut_namespace:
            import nunavut.version
            from nunavut.jinja.loaders import DSDLTemplateLoader

            setattr(nunavut_namespace, "version", nunavut.version.__version__)
            setattr(nunavut_namespace, "platform_version", self._create_platform_version())

            if isinstance(self.loader, DSDLTemplateLoader):
                setattr(nunavut_namespace, "template_sets", self.loader.get_template_sets())



#
# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""Logic for parsing language configuration.

"""
import re
import types
import typing

from yaml import Loader as YamlLoader
from yaml import load as yaml_loader


class LanguageConfig:
    """
    Configuration storage encapsulating parsers and other configuration format details. For any configuration type used
    the concept of "sections" must be maintained which requires that the top-level configuration be structured as
    key/value pairs with the keys using the form "nunavut.lang.[language name]". For example, yaml configuration must
    have a top-level structure like this:

    .. code-block:: python

        example_yaml = '''
            nunavut.lang.a:
                key_one: value_one
                key_two: value_two
            nunavut.lang.b:
                key_one: value_one
                key_two: value_two
            nunavut.lang.c:
                key_one: value_one
                key_two: value_two
        '''

    .. invisible-code-block: python
        from nunavut.lang import LanguageConfig

        config = LanguageConfig()
        config.read_string(example_yaml)

        data = config.sections()
        assert len(data) == 3
        assert data['nunavut.lang.b']['key_two'] == 'value_two'

    .. note::
        The "language name" part of the section identifier must not start with a number and can contain only
        alphanumeric characters. That is, the section identifier must match this pattern:

            nunavut\\.lang\\.[a-zA-Z]{1}\\w*

    The values of the section data can be anything:

    .. code-block:: python

        example_yaml = '''
            nunavut.lang.d:
                key_one:
                    - is
                    - a
                    - list:
                        where: index2
                        is: a_dictionary
        '''

    .. invisible-code-block: python

        config.read_string(example_yaml)
        assert 'a_dictionary' == config.sections()['nunavut.lang.d']['key_one'][2]['list']['is']

    """

    def __init__(self):  # type: ignore
        self._section_name_pattern = re.compile(r"^nunavut\.lang\.([a-zA-Z]{1}\w*)$")
        self._sections = dict()  # type: typing.Dict[str, typing.Dict[str, typing.Any]]

    def update(self, configuration: typing.Any) -> None:
        """
        Add configuration data to this configuration from a string.

        Unlike add_section, this method will update section data with existing keys.

        For example, the first update is to an empty configuration so it will act as a simple insert operation:

        .. invisible-code-block: python
            from nunavut.lang import LanguageConfig

        .. code-block: python

            initial_data = {
                'nunavut.lang.a':
                    {
                        'key_one': 'value_one',
                        'key_two': 'value_two'
                    },
                'nunavut.lang.b':
                    {
                        'key_one':
                            [
                                'item_0',
                                {
                                    'item_1_value_0': 0,
                                    'item_1_value_1': 1
                                },
                                'item_2'
                            ],
                        'key_two': 'value_two'
                    }
            }

            config = LanguageConfig()
            config.update(initial_data)

            assert config.sections()['nunavut.lang.a']['key_one'] == 'value_one'
            assert config.sections()['nunavut.lang.b']['key_one'][1]['item_1_value_1'] == 1
            assert config.sections()['nunavut.lang.b']['key_two'] == 'value_two'

        ...but updating this data is now possible where sections can be added and updated:

        .. code-block: python

            updated_data = '''
                nunavut.lang.b:
                    key_one: simple
                    key_three: value_three
                nunavut.lang.c:
                    key_one: new_language_key
            '''

            updated_data = {
                'nunavut.lang.b':
                    {
                        'key_one': 'simple',
                        'key_three': 'value_three'
                    },
                'nunavut.lang.c':
                    {
                        'key_one': 'new_language_key'
                    },
            }

            config.update(updated_data)

            assert config.sections()['nunavut.lang.a']['key_one'] == 'value_one'
            assert config.sections()['nunavut.lang.b']['key_one'] == 'simple'
            assert config.sections()['nunavut.lang.b']['key_two'] == 'value_two'
            assert config.sections()['nunavut.lang.b']['key_three'] == 'value_three'
            assert config.sections()['nunavut.lang.c']['key_one'] == 'new_language_key'

        .. invisible-code-block: python

            tests = [
                (
                    {
                        'did_not_start.with.nunavut':
                            {
                                'key_one': 'value_one'
                            }
                    },
                    ValueError()
                ),
                (
                    {
                        'nunavut.lang.0c':
                            {
                                'key_one': 'value_one'
                            }
                    },
                    ValueError()
                ),
                (
                    {
                        1:
                            {
                                'key_one': 'value_one'
                            }
                    },
                    TypeError()
                ),
                (
                    {
                        'nunavut.lang.c.iso':
                            {
                                'key_one': 'value_one'
                            }
                    },
                    ValueError()
                )
            ]

            for test_tuple in tests:
                try:
                    config.update(test_tuple[0])
                    assert False
                except BaseException as e:
                    assert isinstance(e, type(test_tuple[1]))
                    pass

        """

        # validate the (very loose) configuration schema.
        for section_name, section_data in configuration.items():
            if not isinstance(section_name, str):
                raise TypeError("section names must be strings")
            if not self._section_name_pattern.match(section_name):
                raise ValueError(
                    'Section name "{}" is invalid. See LanguageConfig documentation for rules.'.format(section_name)
                )
            try:
                section = self._sections[section_name]
            except KeyError:
                self._sections[section_name] = dict()
                section = self._sections[section_name]
            section.update(section_data)

    def sections(self) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        return self._sections

    def read_string(self, string: str, context: typing.Optional[str] = None) -> None:
        configuration = yaml_loader(string, Loader=YamlLoader)
        self.update(configuration)

    def read_file(self, f: typing.TextIO, context: typing.Optional[str] = None) -> None:
        configuration = yaml_loader(f, Loader=YamlLoader)
        self.update(configuration)

    def set(self, section: str, option: str, value: typing.Any) -> None:
        self._sections[section][option] = value

    def add_section(self, section_name: str) -> None:
        if not isinstance(section_name, str):
            raise TypeError("section names must be strings")
        if not self._section_name_pattern.match(section_name):
            raise ValueError(
                'Section name "{}" is invalid. See LanguageConfig documentation for rules.'.format(section_name)
            )
        if section_name in self._sections:
            raise ValueError("Section {} is already defined.".format(section_name))
        self._sections[section_name] = dict()

    _UNSET = object()  # Used internally to allow "None" as a default value.

    def _get_config_value_raw(self, section_name: str, key: str, default_value: typing.Any) -> typing.Any:
        """
        .. invisible-code-block: python
            from nunavut.lang import LanguageConfig

        .. code-block: python

            test_data = {
                'nunavut.lang.a':
                    {
                        'key_one': 'value_one'
                    },
                'nunavut.lang.b':
                    {
                        'key_one': 'value_one'
                    }
            }

            config = LanguageConfig()
            config.update(initial_data)

            try:
                config.get_config_value('nunavut.lang.c', 'foo')
                assert False
            except KeyError:
                pass

            assert 'bar' == config.get_config_value('nunavut.lang.c', 'foo', 'bar')

            try:
                config.get_config_value('nunavut.lang.a', 'foo')
                assert False
            except KeyError:
                pass

            assert 'bar' == config.get_config_value('nunavut.lang.a', 'foo', 'bar')

            assert 'value_one' == config.get_config_value('nunavut.lang.a', 'key_one')
            assert 'value_one' == config.get_config_value('nunavut.lang.a', 'key_one', 'bar')
        """
        try:
            section_data = self._sections[section_name]
        except KeyError:
            if default_value is not self._UNSET:
                return default_value
            else:
                raise
        try:
            return section_data[key]
        except KeyError:
            if default_value is not self._UNSET:
                return default_value
            else:
                raise

    def get_config_value(self, section_name: str, key: str, default_value: typing.Optional[str] = None) -> str:
        """
        Get an optional language property from the language configuration.

        :param section_name : The name of the section to get the value from.
        :param str key      : The config value to retrieve.
        :param default_value: The value to return if the key was not in the configuration. If provided
            this method will not raise.
        :type default_value : typing.Optional[str]
        :return: Either the value from the config or the default_value if provided.
        :rtype: str
        :raises: KeyError if the section or the key in the section does not exist and a default_value was not provided.

         .. invisible-code-block: python
            from nunavut.lang import LanguageConfig

        .. code-block: python

            test_data = {
                'nunavut.lang.a':
                    {
                        'key_one': [
                            1,
                            2
                        ]
                    },
                'nunavut.lang.b':
                    {
                        'key_one': 'value_one'
                    }
            }

            config = LanguageConfig()
            config.update(test_data)

            assert 'value_one' == config.get_config_value('nunavut.lang.b', 'key_one')
            assert 'value_one' == config.get_config_value('nunavut.lang.b', 'key_one', 'bar')
            assert 'bar' == config.get_config_value('nunavut.lang.b', 'key_two', 'bar')
            try:
                config.get_config_value('nunavut.lang.b', 'key_two')
                assert False # supposed to throw without a default value.
            except KeyError:
                pass
        """
        optional_result = self._get_config_value_raw(
            section_name, key, (self._UNSET if default_value is None else default_value)
        )

        # when we've retrieved a None result the str() cast will return "None" which isn't our intent.
        # Instead, if we get None and the _get_config_value_raw didn't throw a KeyError then what we really
        # meant is that we wanted an empty string if the value existed but was None.
        return str(optional_result) if optional_result is not None else ""

    def get_config_value_as_bool(self, section_name: str, key: str, default_value: bool = False) -> bool:
        """
        Get an optional language property from the language configuration returning a boolean. The rules
        for boolean conversion are as follows:

        .. invisible-code-block: python

            from nunavut.lang import LanguageConfig

            config = LanguageConfig()
            config.add_section('nunavut.lang.cpp')

        .. code-block:: python

            # "Any string" = True
            config.set('nunavut.lang.cpp', 'v', 'Any string')
            assert config.get_config_value_as_bool('nunavut.lang.cpp', 'v')

            # "true" = True
            config.set('nunavut.lang.cpp', 'v', 'true')
            assert config.get_config_value_as_bool('nunavut.lang.cpp', 'v')

            # "TrUe" = True
            config.set('nunavut.lang.cpp', 'v', 'TrUe')
            assert config.get_config_value_as_bool('nunavut.lang.cpp', 'v')

            # "1" = True
            config.set('nunavut.lang.cpp', 'v', '1')
            assert config.get_config_value_as_bool('nunavut.lang.cpp', 'v')

            # "false" = False
            config.set('nunavut.lang.cpp', 'v', 'false')
            assert not config.get_config_value_as_bool('nunavut.lang.cpp', 'v')

            # "FaLse" = False
            config.set('nunavut.lang.cpp', 'v', 'FaLse')
            assert not config.get_config_value_as_bool('nunavut.lang.cpp', 'v')

            # "0" = False
            config.set('nunavut.lang.cpp', 'v', '0')
            assert not config.get_config_value_as_bool('nunavut.lang.cpp', 'v')

            # "" = False
            config.set('nunavut.lang.cpp', 'v', '')
            assert not config.get_config_value_as_bool('nunavut.lang.cpp', 'v')

            # False if not defined
            assert not config.get_config_value_as_bool('nunavut.lang.cpp', 'not_a_key')

            # True if not defined but default_value is True
            assert not config.get_config_value_as_bool('nunavut.lang.cpp', 'not_a_key')

        :param section_name         : The name of the section to get the value from.
        :param str key              : The config value to retrieve.
        :param bool default_value   : The value to use if no value existed.
        :return                     : The config value as either True or False.
        :rtype                      : bool
        """
        result = self.get_config_value(section_name, key, default_value="false" if not default_value else "true")
        if result.lower() == "false" or result == "0":
            return False
        else:
            return bool(result)

    def get_config_value_as_dict(
        self, section_name: str, key: str, default_value: typing.Optional[typing.Dict] = None
    ) -> typing.Dict[str, typing.Any]:
        """
        Get a language property parsing it as a map with string keys.

        Example:

        .. invisible-code-block: python

            from nunavut.lang import LanguageConfig

            config = LanguageConfig()
            config.add_section('nunavut.lang.a')

        .. code-block: python

            config.set('nunavut.lang.a', 'foo', {'one': 1})

            assert config.get_config_value_as_dict('nunavut.lang.a', 'foo')['one'] == 1

        .. invisible-code-block: python

            try:
                config.get_config_value_as_dict('nunavut.lang.b', 'foo')
                assert False
            except KeyError:
                pass

            assert config.get_config_value_as_dict('nunavut.lang.b', 'foo', {'one': 2})['one'] == 2

            try:
                config.get_config_value_as_dict('nunavut.lang.a', 'bar')
                assert False
            except KeyError:
                pass

            assert config.get_config_value_as_dict('nunavut.lang.a', 'bar', {'one': 2})['one'] == 2

            config.set('nunavut.lang.a', 'bar', 1)

            try:
                config.get_config_value_as_dict('nunavut.lang.a', 'bar')
                assert False
            except TypeError:
                pass

            assert config.get_config_value_as_dict('nunavut.lang.a', 'bar', {'one': 2})['one'] == 2

        :param str section_name : The name of the section to get the key from.
        :param str key          : The config value to retrieve.
        :param default_value    : The value to return if the key was not in the configuration. If provided this method
            will not raise a KeyError nor a TypeError.
        :type default_value     : typing.Optional[typing.Mapping[str, typing.Any]]
        :return                 : Either the value from the config or the default_value if provided.
        :rtype                  : typing.Mapping[str, typing.Any]
        :raises                 : KeyError if the key does not exist and a default_value was not provided.
        :raises                 : TypeError if the value exists but is not a dict and a default_value was not provided.

        """
        raw_value = self._get_config_value_raw(
            section_name, key, default_value=(self._UNSET if default_value is None else default_value)
        )
        if isinstance(raw_value, dict):
            return raw_value

        if default_value is None:
            raise TypeError("{}.{} exists but is not a dict. (is type {})".format(section_name, key, type(raw_value)))

        return default_value

    def get_config_value_as_list(
        self, section_name: str, key: str, default_value: typing.Optional[typing.List] = None
    ) -> typing.List[typing.Any]:
        """
        Get a language property parsing it as a map with string keys.

        Example:

        .. invisible-code-block: python

            from nunavut.lang import LanguageConfig

            config = LanguageConfig()
            config.add_section('nunavut.lang.a')

        .. code-block: python

            config.set('nunavut.lang.a', 'foo', [1, 2, 3])

            assert config.get_config_value_as_list('nunavut.lang.a', 'foo')[1] == 2

        .. invisible-code-block: python

            try:
                config.get_config_value_as_list('nunavut.lang.b', 'foo')
                assert False
            except KeyError:
                pass

            assert config.get_config_value_as_list('nunavut.lang.b', 'foo', [2, 3, 4])[1] == 3

            try:
                config.get_config_value_as_list('nunavut.lang.a', 'bar')
                assert False
            except KeyError:
                pass

            assert config.get_config_value_as_list('nunavut.lang.a', 'bar', [3, 4, 5])[1] == 4

            config.set('nunavut.lang.a', 'bar', 1)

            try:
                config.get_config_value_as_list('nunavut.lang.a', 'bar')
                assert False
            except TypeError:
                pass

            assert config.get_config_value_as_list('nunavut.lang.a', 'bar', [4, 5, 6])[1] == 5

        :param str section_name : The name of the section to get the key from.
        :param str key          : The config value to retrieve.
        :param default_value    : The value to return if the key was not in the configuration. If provided this method
            will not raise a KeyError nor a TypeError.
        :type default_value     : typing.Optional[typing.List[typing.Any]]
        :return                 : Either the value from the config or the default_value if provided.
        :rtype                  : typing.List[typing.Any]
        :raises                 : KeyError if the key does not exist and a default_value was not provided.
        :raises                 : TypeError if the value exists but is not a dict and a default_value was not provided.

        """
        raw_value = self._get_config_value_raw(
            section_name, key, default_value=(self._UNSET if default_value is None else default_value)
        )
        if isinstance(raw_value, list):
            return raw_value

        if default_value is None:
            raise TypeError("{}.{} exists but is not a list. (is type {})".format(section_name, key, type(raw_value)))

        return default_value


# +-------------------------------------------------------------------------------------------------------------------+
# | VersionReader
# +-------------------------------------------------------------------------------------------------------------------+


class VersionReader:
    """
    Helper to read an "x.y.z" semantic version from python modules as a module variable
    "__version__"
    """

    MODULE_VERSION_ATTRIBUTE_NAME = "__version__"

    @classmethod
    def parse_version(cls, version_string: str) -> typing.Optional[typing.Tuple[int, int, int]]:
        version_array = [int(x) for x in version_string.split(".")]
        if len(version_array) != 3:
            return None
        else:
            return (version_array[0], version_array[1], version_array[2])

    @classmethod
    def read_version(cls, module: "types.ModuleType") -> typing.Tuple[int, int, int]:
        version = getattr(module, cls.MODULE_VERSION_ATTRIBUTE_NAME, "0.0.0")  # type: str

        version_tuple = cls.parse_version(version)
        if version_tuple is None:
            raise RuntimeError(
                'Invalid {} "{}" for module {} (expected "x.y.z")'.format(
                    cls.MODULE_VERSION_ATTRIBUTE_NAME, version, module.__name__
                )
            )
        return version_tuple

    def __init__(self, module_name: str):
        self._module_name = module_name
        self._cached = None  # type: typing.Optional[typing.Tuple[int, int, int]]

    @property
    def version(self) -> typing.Tuple[int, int, int]:
        if self._cached is None:
            self._cached = self._get_version()
        return self._cached

    def _get_version(self) -> typing.Tuple[int, int, int]:
        import importlib

        try:
            return self.read_version(importlib.import_module(self._module_name))
        except (ImportError, ValueError):
            return (0, 0, 0)



from nunavut.jinja.jinja2.visitor import NodeVisitor
from nunavut.jinja.jinja2._compat import iteritems


VAR_LOAD_PARAMETER = 'param'
VAR_LOAD_RESOLVE = 'resolve'
VAR_LOAD_ALIAS = 'alias'
VAR_LOAD_UNDEFINED = 'undefined'


def find_symbols(nodes, parent_symbols=None):
    sym = Symbols(parent=parent_symbols)
    visitor = FrameSymbolVisitor(sym)
    for node in nodes:
        visitor.visit(node)
    return sym


def symbols_for_node(node, parent_symbols=None):
    sym = Symbols(parent=parent_symbols)
    sym.analyze_node(node)
    return sym


class Symbols(object):

    def __init__(self, parent=None, level=None):
        if level is None:
            if parent is None:
                level = 0
            else:
                level = parent.level + 1
        self.level = level
        self.parent = parent
        self.refs = {}
        self.loads = {}
        self.stores = set()

    def analyze_node(self, node, **kwargs):
        visitor = RootVisitor(self)
        visitor.visit(node, **kwargs)

    def _define_ref(self, name, load=None):
        ident = 'l_%d_%s' % (self.level, name)
        self.refs[name] = ident
        if load is not None:
            self.loads[ident] = load
        return ident

    def find_load(self, target):
        if target in self.loads:
            return self.loads[target]
        if self.parent is not None:
            return self.parent.find_load(target)

    def find_ref(self, name):
        if name in self.refs:
            return self.refs[name]
        if self.parent is not None:
            return self.parent.find_ref(name)

    def ref(self, name):
        rv = self.find_ref(name)
        if rv is None:
            raise AssertionError('Tried to resolve a name to a reference that '
                                 'was unknown to the frame (%r)' % name)
        return rv

    def copy(self):
        rv = object.__new__(self.__class__)
        rv.__dict__.update(self.__dict__)
        rv.refs = self.refs.copy()
        rv.loads = self.loads.copy()
        rv.stores = self.stores.copy()
        return rv

    def store(self, name):
        self.stores.add(name)

        # If we have not see the name referenced yet, we need to figure
        # out what to set it to.
        if name not in self.refs:
            # If there is a parent scope we check if the name has a
            # reference there.  If it does it means we might have to alias
            # to a variable there.
            if self.parent is not None:
                outer_ref = self.parent.find_ref(name)
                if outer_ref is not None:
                    self._define_ref(name, load=(VAR_LOAD_ALIAS, outer_ref))
                    return

            # Otherwise we can just set it to undefined.
            self._define_ref(name, load=(VAR_LOAD_UNDEFINED, None))

    def declare_parameter(self, name):
        self.stores.add(name)
        return self._define_ref(name, load=(VAR_LOAD_PARAMETER, None))

    def load(self, name):
        target = self.find_ref(name)
        if target is None:
            self._define_ref(name, load=(VAR_LOAD_RESOLVE, name))

    def branch_update(self, branch_symbols):
        stores = {}
        for branch in branch_symbols:
            for target in branch.stores:
                if target in self.stores:
                    continue
                stores[target] = stores.get(target, 0) + 1

        for sym in branch_symbols:
            self.refs.update(sym.refs)
            self.loads.update(sym.loads)
            self.stores.update(sym.stores)

        for name, branch_count in iteritems(stores):
            if branch_count == len(branch_symbols):
                continue
            target = self.find_ref(name)
            assert target is not None, 'should not happen'

            if self.parent is not None:
                outer_target = self.parent.find_ref(name)
                if outer_target is not None:
                    self.loads[target] = (VAR_LOAD_ALIAS, outer_target)
                    continue
            self.loads[target] = (VAR_LOAD_RESOLVE, name)

    def dump_stores(self):
        rv = {}
        node = self
        while node is not None:
            for name in node.stores:
                if name not in rv:
                    rv[name] = self.find_ref(name)
            node = node.parent
        return rv

    def dump_param_targets(self):
        rv = set()
        node = self
        while node is not None:
            for target, (instr, _) in iteritems(self.loads):
                if instr == VAR_LOAD_PARAMETER:
                    rv.add(target)
            node = node.parent
        return rv


class RootVisitor(NodeVisitor):

    def __init__(self, symbols):
        self.sym_visitor = FrameSymbolVisitor(symbols)

    def _simple_visit(self, node, **kwargs):
        for child in node.iter_child_nodes():
            self.sym_visitor.visit(child)

    visit_Template = visit_Block = visit_Macro = visit_FilterBlock = \
        visit_Scope = visit_If = visit_ScopedEvalContextModifier = \
        _simple_visit

    def visit_AssignBlock(self, node, **kwargs):
        for child in node.body:
            self.sym_visitor.visit(child)

    def visit_CallBlock(self, node, **kwargs):
        for child in node.iter_child_nodes(exclude=('call',)):
            self.sym_visitor.visit(child)

    def visit_OverlayScope(self, node, **kwargs):
        for child in node.body:
            self.sym_visitor.visit(child)

    def visit_For(self, node, for_branch='body', **kwargs):
        if for_branch == 'body':
            self.sym_visitor.visit(node.target, store_as_param=True)
            branch = node.body
        elif for_branch == 'else':
            branch = node.else_
        elif for_branch == 'test':
            self.sym_visitor.visit(node.target, store_as_param=True)
            if node.test is not None:
                self.sym_visitor.visit(node.test)
            return
        else:
            raise RuntimeError('Unknown for branch')
        for item in branch or ():
            self.sym_visitor.visit(item)

    def visit_With(self, node, **kwargs):
        for target in node.targets:
            self.sym_visitor.visit(target)
        for child in node.body:
            self.sym_visitor.visit(child)

    def generic_visit(self, node, *args, **kwargs):
        raise NotImplementedError('Cannot find symbols for %r' %
                                  node.__class__.__name__)


class FrameSymbolVisitor(NodeVisitor):
    """A visitor for `Frame.inspect`."""

    def __init__(self, symbols):
        self.symbols = symbols

    def visit_Name(self, node, store_as_param=False, **kwargs):
        """All assignments to names go through this function."""
        if store_as_param or node.ctx == 'param':
            self.symbols.declare_parameter(node.name)
        elif node.ctx == 'store':
            self.symbols.store(node.name)
        elif node.ctx == 'load':
            self.symbols.load(node.name)

    def visit_NSRef(self, node, **kwargs):
        self.symbols.load(node.name)

    def visit_If(self, node, **kwargs):
        self.visit(node.test, **kwargs)

        original_symbols = self.symbols

        def inner_visit(nodes):
            self.symbols = rv = original_symbols.copy()
            for subnode in nodes:
                self.visit(subnode, **kwargs)
            self.symbols = original_symbols
            return rv

        body_symbols = inner_visit(node.body)
        elif_symbols = inner_visit(node.elif_)
        else_symbols = inner_visit(node.else_ or ())

        self.symbols.branch_update([body_symbols, elif_symbols, else_symbols])

    def visit_Macro(self, node, **kwargs):
        self.symbols.store(node.name)

    def visit_Import(self, node, **kwargs):
        self.generic_visit(node, **kwargs)
        self.symbols.store(node.target)

    def visit_FromImport(self, node, **kwargs):
        self.generic_visit(node, **kwargs)
        for name in node.names:
            if isinstance(name, tuple):
                self.symbols.store(name[1])
            else:
                self.symbols.store(name)

    def visit_Assign(self, node, **kwargs):
        """Visit assignments in the correct order."""
        self.visit(node.node, **kwargs)
        self.visit(node.target, **kwargs)

    def visit_For(self, node, **kwargs):
        """Visiting stops at for blocks.  However the block sequence
        is visited as part of the outer scope.
        """
        self.visit(node.iter, **kwargs)

    def visit_CallBlock(self, node, **kwargs):
        self.visit(node.call, **kwargs)

    def visit_FilterBlock(self, node, **kwargs):
        self.visit(node.filter, **kwargs)

    def visit_With(self, node, **kwargs):
        for target in node.values:
            self.visit(target)

    def visit_AssignBlock(self, node, **kwargs):
        """Stop visiting at block assigns."""
        self.visit(node.target, **kwargs)

    def visit_Scope(self, node, **kwargs):
        """Stop visiting at scopes."""

    def visit_Block(self, node, **kwargs):
        """Stop visiting at blocks."""

    def visit_OverlayScope(self, node, **kwargs):
        """Do not visit into overlay scopes."""



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Command-line entrypoint.
"""

import sys

from .cli import main

sys.exit(main())



#!/usr/bin/env python3
#
# Copyright (C) 2018  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import sys
import setuptools

from typing import Dict

if int(setuptools.__version__.split('.')[0]) < 30:
    print('A newer version of setuptools is required. The current version does not support declarative config.',
          file=sys.stderr)
    sys.exit(1)

version = {}  # type: Dict
with open('src/nunavut/version.py') as fp:
    exec(fp.read(), version)

setuptools.setup(version=version['__version__'],
                 package_data={'': ['*.j2', '**/*.css', '**/*.js', '*.ini', '*.yaml', '*.hpp', '*.h']})



#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""Language-specific support in nunavut.

This package contains modules that provide specific support for generating
source for various languages using templates.
"""
import abc
import functools
import importlib
import logging
import pathlib
import types
import typing

import pydsdl

from ..dependencies import Dependencies, DependencyBuilder
from .._utilities import YesNoDefault, iter_package_resources
from ._config import LanguageConfig, VersionReader

logger = logging.getLogger(__name__)


class LanguageLoader:
    """
    Factory class that loads language meta-data and concrete :class:`nunavut.lang.Language` objects.

    :param additional_config_files: A list of paths to additional configuration files to load as configuration.
        These will override any values found in the :file:`nunavut.lang.properties.yaml` file and files
        appearing later in this list will override value found in earlier entries.

        .. invisible-code-block: python

            from nunavut.lang import LanguageLoader

            subject = LanguageLoader()

            c_reserved_identifiers = subject.config.get_config_value('nunavut.lang.c','reserved_identifiers')
            assert len(c_reserved_identifiers) > 0
    """

    @classmethod
    def load_language_module(cls, language_name: str) -> "types.ModuleType":
        module_name = "nunavut.lang.{}".format(language_name)
        return importlib.import_module(module_name)

    @classmethod
    def _load_config(cls, *additional_config_files: pathlib.Path) -> LanguageConfig:
        parser = LanguageConfig()
        for resource in iter_package_resources(__name__, ".yaml"):
            ini_string = resource.read_text()
            parser.read_string(ini_string)
        for additional_path in additional_config_files:
            with open(str(additional_path), "r") as additional_file:
                parser.read_file(additional_file)
        return parser

    def __init__(self, *additional_config_files: pathlib.Path):
        self._config = None  # type: typing.Optional[LanguageConfig]
        self._additional_config_files = additional_config_files

    @property
    def config(self) -> LanguageConfig:
        """
        Meta-data about all languages merged from the Nunavut internal defaults and any additional
        configuration files provided to this class's constructor.
        """
        if self._config is None:
            self._config = self._load_config(*self._additional_config_files)
        return self._config

    def load_language(
        self,
        language_name: str,
        omit_serialization_support: bool,
        language_options: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    ) -> "Language":
        """
        :param str language_name:                The name of the language used by the :mod:`nunavut.lang` module.
        :param LanguageConfig config:            The parser to load language properties into.
        :param bool omit_serialization_support:  The value to set for the :func:`omit_serialization_support` property
                                                for this language.
        :param typing.Optional[typing.Mapping[str, typing.Any]] language_options: Opaque arguments passed through to the
                    target :class:`nunavut.lang.Language` object.

        :return: A new object that extends :class:`nunavut.lang.Language`.
        :rtype: nunavut.lang.Language

        .. invisible-code-block: python

            from nunavut.lang import LanguageLoader

        .. code-block:: python

            lang_c = LanguageLoader().load_language('c', True)
            assert lang_c.name == 'c'

        .. invisible-code-block: python

            # let's go ahead and load the rest of our known, internally supported languages just to raise
            # test failures right here at the wellspring.

            lang_cpp = LanguageLoader().load_language('py', True)
            assert lang_cpp.name == 'py'

            lang_js = LanguageLoader().load_language('js', True)
            assert lang_js.name == 'js'

        """
        ln_module = self.load_language_module(language_name)

        try:
            language_type = typing.cast(typing.Type["Language"], getattr(ln_module, "Language"))
        except AttributeError:
            logging.debug(
                "Unable to find a Language object in nunavut.lang.{}. Using a Generic language object".format(
                    language_name
                )
            )
            language_type = _GenericLanguage

        if language_type == Language:
            # the language module just imported the base class so let's go ahead and use _GenericLanguage
            language_type = _GenericLanguage

        return language_type(ln_module, self.config, omit_serialization_support, language_options)


class Language(metaclass=abc.ABCMeta):
    """
    Facilities for generating source code for a specific language. Concrete Language classes must be implemented
    by the language support package below lang and should be instantiated using
    :class:`nunavut.lang.LanguageLoader`.

    :param str language_name:                The name of the language used by the :mod:`nunavut.lang` module.
    :param LanguageConfig config:            The parser to load language properties into.
    :param bool omit_serialization_support:  The value to set for the :func:`omit_serialization_support` property
                                             for this language.
    :param typing.Optional[typing.Mapping[str, typing.Any]] language_options: Opaque arguments passed through to the
                target :class:`nunavut.lang.Language` object.

        .. invisible-code-block: python

            from nunavut.lang import Language, _GenericLanguage
            from unittest.mock import MagicMock

            mock_config = MagicMock()
            mock_module = MagicMock()

            mock_module.__name__ = 'foo'
            try:
                my_lang = _GenericLanguage(mock_module, mock_config, True)
                # module must be within 'nunavut'
                assert False
            except RuntimeError:
                pass

            mock_module.__name__ = 'nunavut.foo'
            try:
                my_lang = _GenericLanguage(mock_module, mock_config, True)
                # module must be within 'nunavut.lang'
                assert False
            except RuntimeError:
                pass

            mock_module.__name__ = 'not.nunavut.foo'
            try:
                my_lang = _GenericLanguage(mock_module, mock_config, True)
                # module must be within 'nunavut.lang'
                assert False
            except RuntimeError:
                pass

            mock_module.__name__ = 'nunavut.lang.foo'
            my_lang = _GenericLanguage(mock_module, mock_config, True)
            assert my_lang.name == 'foo'
    """

    @classmethod
    def default_filter_id_for_target(cls, instance: typing.Any) -> str:
        """
        The default transformation of any object into a string.

        :param any instance:        Any object or data that either has a name property or can be converted to a string.
        :return: Either ``str(instance.name)`` if the instance has a name property or just ``str(instance)``
        """
        if hasattr(instance, "name"):
            return str(instance.name)
        else:
            return str(instance)

    def __init__(
        self,
        language_module: "types.ModuleType",
        config: LanguageConfig,
        omit_serialization_support: bool,
        language_options: typing.Optional[typing.Mapping[str, typing.Any]] = None,
    ):
        self._globals = None  # type: typing.Optional[typing.Mapping[str, typing.Any]]
        self._section = language_module.__name__
        name_parts = self._section.split(".")
        if len(name_parts) != 3 or name_parts[0] != "nunavut" or name_parts[1] != "lang":
            raise RuntimeError("unknown module provided to Language class.")
        self._language_name = name_parts[2]
        self._config = config
        self._omit_serialization_support = omit_serialization_support
        self._language_options = config.get_config_value_as_dict(self._section, "options", dict())

        if language_options is not None:
            self._language_options.update(language_options)

        self._filters = dict()  # type: typing.Dict[str, typing.Callable]
        self._tests = dict()  # type: typing.Dict[str, typing.Callable]
        self._uses = dict()  # type: typing.Dict[str, typing.Callable]

    def __getattr__(self, name: str) -> typing.Any:
        """
        Any attribute access to a Language object will return the regular properties and
        any globals defined for the language. Because of this do not extend properties
        on this object in a way that will clash with the globals it defines (e.g. typename_
        or valuetoken_ should not be used as the start of attribute names).
        """
        try:
            return self.get_globals()[name]
        except KeyError as e:
            raise AttributeError(e)

    def get_support_module(self) -> typing.Tuple[str, typing.Tuple[int, int, int], typing.Optional["types.ModuleType"]]:
        """
        Returns the module object for the language support files.
        :return: A tuple of module name, x.y.z module version, and the module object itself.

        .. invisible-code-block: python

            from nunavut.lang import Language, _GenericLanguage
            from unittest.mock import MagicMock

            mock_config = MagicMock()
            mock_module = MagicMock()
            mock_module.__name__ = 'nunavut.lang.cpp'

            my_lang = _GenericLanguage(mock_module, mock_config, True)
            my_lang._section = "nunavut.lang.cpp"
            module_name, support_version, _ = my_lang.get_support_module()

            assert module_name == "nunavut.lang.cpp.support"
            assert support_version[0] == 1

        """
        module_name = "{}.support".format(self._section)

        try:
            module = importlib.import_module(module_name)
            version_tuple = VersionReader.read_version(module)
            return (module_name, version_tuple, module)
        except (ImportError, ValueError):
            # No serialization support for this language
            logger.info("No serialization support for selected target. Cannot retrieve module.")

            return (module_name, (0, 0, 0), None)

    @functools.lru_cache()
    def get_dependency_builder(self, for_type: pydsdl.Any) -> DependencyBuilder:
        return DependencyBuilder(for_type)

    @abc.abstractmethod
    def get_includes(self, dep_types: Dependencies) -> typing.List[str]:
        """
        Get a list of include paths that are specific to this language and the options set for it.
        :param Dependencies dep_types: A description of the dependencies includes are needed for.
        :return: A list of include file paths. The list may be empty if no includes were needed.
        """
        pass

    def filter_id(self, instance: typing.Any, id_type: str = "any") -> str:
        """
        Produces a valid identifier in the language for a given object. The encoding may not be reversible.

        :param any instance:        Any object or data that either has a name property or can be converted
                                    to a string.
        :param str id_type:         A type of identifier. This is different for each language. For example, for C this
                                    value can be 'typedef', 'macro', 'function', or 'enum'.
                                    Use 'any' to apply stropping rules for all identifier types to the instance.
        :return: A token that is a valid identifier in the language, is not a reserved keyword, and is transformed
                in a deterministic manner based on the provided instance.
        """
        return self.default_filter_id_for_target(instance)

    def filter_short_reference_name(
        self, t: pydsdl.CompositeType, stropping: YesNoDefault = YesNoDefault.DEFAULT, id_type: str = "any"
    ) -> str:
        """
        Provides a string that is a shorted version of the full reference name omitting any namespace parts of the type.

        :param pydsdl.CompositeType t: The DSDL type to get the reference name for.
        :param YesNoDefault stropping: If DEFAULT then the stropping value configured for the target language is used
                                       else this overrides that value.
        :param str id_type:         A type of identifier. This is different for each language. For example, for C this
                                    value can be 'typedef', 'macro', 'function', or 'enum'.
                                    Use 'any' to apply stropping rules for all identifier types to the instance.
        """
        short_name = "{short}_{major}_{minor}".format(short=t.short_name, major=t.version.major, minor=t.version.minor)
        if YesNoDefault.test_truth(stropping, self.enable_stropping):
            return self.filter_id(short_name, id_type)
        else:
            return short_name

    def get_config_value(self, key: str, default_value: typing.Optional[str] = None) -> str:
        """
        Get an optional language property from the language configuration.

        :param str key: The config value to retrieve.
        :param default_value: The value to return if the key was not in the configuration. If provided
                this method will not raise.
        :type default_value: typing.Optional[str]
        :return: Either the value from the config or the default_value if provided.
        :rtype: str
        :raises: KeyError if the section or the key in the section does not exist and a default_value was not provided.
        """
        return self._config.get_config_value(self._section, key, default_value)

    def get_config_value_as_bool(self, key: str, default_value: bool = False) -> bool:
        """
        Get an optional language property from the language configuration returning a boolean.

        :param str key: The config value to retrieve.
        :param bool default_value: The value to use if no value existed.
        :return: The config value as either True or False.
        :rtype: bool
        """
        return self._config.get_config_value_as_bool(self._section, key, default_value)

    def get_config_value_as_dict(
        self, key: str, default_value: typing.Optional[typing.Dict] = None
    ) -> typing.Dict[str, typing.Any]:
        """
        Get a language property parsing it as a map with string keys.

        .. invisible-code-block: python

            from nunavut.lang import LanguageConfig, LanguageLoader, Language, _GenericLanguage

            config = LanguageConfig()
            config.add_section('nunavut.lang.c')
            config.set('nunavut.lang.c', 'foo', {'one': 1})

            lang_c = _GenericLanguage(LanguageLoader.load_language_module('c'), config, True)

            assert lang_c.get_config_value_as_dict('foo')['one'] == 1

            assert lang_c.get_config_value_as_dict('bar', {'one': 2})['one'] == 2

        :param str key:           The config value to retrieve.
        :param default_value:     The value to return if the key was not in the configuration. If provided this method
            will not raise a KeyError nor a TypeError.
        :type default_value: typing.Optional[typing.Mapping[str, typing.Any]]
        :return:                  Either the value from the config or the default_value if provided.
        :rtype: typing.Mapping[str, typing.Any]
        :raises: KeyError if the key does not exist and a default_value was not provided.
        :raises: TypeError if the value exists but is not a dict and a default_value was not provided.

        """
        return self._config.get_config_value_as_dict(self._section, key, default_value)

    def get_config_value_as_list(
        self, key: str, default_value: typing.Optional[typing.List] = None
    ) -> typing.List[typing.Any]:
        """
        Get a language property parsing it as a map with string keys.

        :param str key:           The config value to retrieve.
        :param default_value:     The value to return if the key was not in the configuration. If provided this method
            will not raise a KeyError nor a TypeError.
        :type default_value:      typing.Optional[typing.List[typing.Any]]
        :return:                  Either the value from the config or the default_value if provided.
        :rtype:                   typing.List[typing.Any]
        :raises:                  KeyError if the key does not exist and a default_value was not provided.
        :raises:                  TypeError if the value exists but is not a dict and a default_value was not provided.

        """
        return self._config.get_config_value_as_list(self._section, key, default_value)

    @property
    def extension(self) -> str:
        """
        The extension to use for files generated in this language.
        """
        return self._config.get_config_value(self._section, "extension", "get")

    @property
    def namespace_output_stem(self) -> typing.Optional[str]:
        """
        The name of a namespace file for this language.
        """
        try:
            return self._config.get_config_value(self._section, "namespace_file_stem")
        except KeyError:
            return None

    @property
    def name(self) -> str:
        """
        The name of the language used by the :mod:`nunavut.lang` module.
        """
        return self._language_name

    @property
    def support_namespace(self) -> typing.List[str]:
        """
        The hierarchical namespace used by the support software. The property
        is a dot separated string when specified in configuration. This
        property returns that value split into namespace components with the
        first identifier being the first index in the array, etc.

        .. invisible-code-block: python

            from nunavut.lang import Language

            config = {
                'nunavut.lang.cpp':
                {
                    'support_namespace': 'foo.bar'
                }
            }
            lctx = configurable_language_context_factory(config, 'cpp')
            lang_cpp = lctx.get_target_language()

            assert len(lang_cpp.support_namespace) == 2
            assert lang_cpp.support_namespace[0] == 'foo'
            assert lang_cpp.support_namespace[1] == 'bar'

        """
        namespace_str = self._config.get_config_value(self._section, "support_namespace", default_value="")
        return namespace_str.split(".")

    @property
    def enable_stropping(self) -> bool:
        """
        Whether or not to strop identifiers for this language.
        """
        return self._config.get_config_value_as_bool(self._section, "enable_stropping")

    @property
    def has_standard_namespace_files(self) -> bool:
        """
        Whether or not the language defines special namespace files as part of
        its core standard (e.g. python's __init__).
        """
        return self._config.get_config_value_as_bool(self._section, "has_standard_namespace_files")

    @property
    def stable_support(self) -> bool:
        """
        Whether support for this language is designated 'stable', and not experimental.
        """
        return self._config.get_config_value_as_bool(self._section, "stable_support")

    @property
    def omit_serialization_support(self) -> bool:
        """
        If True then generators should not include serialization routines, types,
        or support libraries for this language.
        """
        return self._omit_serialization_support

    @property
    def support_files(self) -> typing.Generator[pathlib.Path, None, None]:
        """
        Iterates over non-templated supporting files embedded within the Nunavut distribution.

        .. invisible-code-block: python

            from nunavut.lang import Language, _GenericLanguage
            from unittest.mock import MagicMock

            mock_config = MagicMock()
            mock_module = MagicMock()
            mock_module.__name__ = 'nunavut.lang.foo'

            my_lang = _GenericLanguage(mock_module, mock_config, True)
            my_lang._section = "nunavut.lang.not_a_language_really_not_a_language"
            for support_file in my_lang.support_files:
                # if the module doesn't exist it shouldn't have any support files.
                assert False

        """
        _, _, module = self.get_support_module()

        if module is not None:
            # All language support modules must provide a list_support_files method
            # to allow the copy generator access to the packaged support files.
            list_support_files = getattr(
                module, "list_support_files"
            )  # type: typing.Callable[[], typing.Generator[pathlib.Path, None, None]]
            return list_support_files()
        else:
            # No serialization support for this language
            def list_support_files() -> typing.Generator[pathlib.Path, None, None]:
                # This makes both MyPy and sonarqube happy.
                return typing.cast(typing.Generator[pathlib.Path, None, None], iter(()))

            return list_support_files()

    def get_option(
        self, option_key: str, default_value: typing.Union[typing.Mapping[str, typing.Any], str, None] = None
    ) -> typing.Union[typing.Mapping[str, typing.Any], str, None]:
        """
        Get a language option for this language.

        .. invisible-code-block: python

            config = {
                'nunavut.lang.cpp':
                {
                    'options': {'target_endianness': 'little'}
                }
            }
            lctx = configurable_language_context_factory(config, 'cpp')
            lang_cpp = lctx.get_target_language()

        .. code-block:: python

            # Values can come from defaults...
            assert lang_cpp.get_option('target_endianness') == 'little'

            # ... or can come from a sane default.
            assert lang_cpp.get_option('foobar', 'sane_default') == 'sane_default'

        :return: Either the value provided to the :class:`Language` instance, the value from properties.yaml,
            or the :code:`default_value`.

        """
        try:
            return self._language_options[option_key]  # type: ignore
        except KeyError:
            return default_value

    def get_templates_package_name(self) -> str:
        """
        The name of the nunavut python package containing filters, types, and configuration
        for this language.
        """
        return self._section

    def get_named_types(self) -> typing.Mapping[str, str]:
        """
        Get a map of named types to the type name to emit for this language.
        """
        return self._config.get_config_value_as_dict(self._section, "named_types", default_value={})

    def get_named_values(self) -> typing.Mapping[str, str]:
        """
        Get a map of named values to the token to emit for this language.
        """
        return self._config.get_config_value_as_dict(self._section, "named_values", default_value={})

    def get_globals(self) -> typing.Mapping[str, typing.Any]:
        """
        Get all values for this language that should be available in a global context.

        :return: A mapping of global names to global values.
        """
        if self._globals is None:
            globals_map = dict()  # type: typing.Dict[str, typing.Any]

            for key, value in self.get_named_types().items():
                globals_map["typename_{}".format(key)] = value
            for key, value in self.get_named_values().items():
                globals_map["valuetoken_{}".format(key)] = value

            self._globals = globals_map
        return self._globals

    def get_options(self) -> typing.Mapping[str, typing.Any]:
        """
        Get all language options for this Language.

        :return: A mapping of option names to option values.
        """
        return self._language_options


class _GenericLanguage(Language):
    """
    Language type used when the language support within Nunavut does not define a language-specific
    subclass.

    Do not use this. Use :class:`nunavut.lang.LanguageLoader` which will create the proper object type
    based on inspection of the Nunavut internals.
    """

    def get_includes(self, dep_types: Dependencies) -> typing.List[str]:
        return []


class LanguageContext:
    """
    Context object containing the current target language (if any) and used to access
    :class:`Language` objects.

    :param str target_language: If provided a :class:`Language` object will be created to hold
                                the target language set for this context. If None then there is
                                no target language.
    :param str extension: The extension to use for generated file types or None to use a default
                          based on the target_language.
    :param str namespace_output_stem: The filename stem to give to Namespace output files if
                                      emitted or None to use a default based on a target_language.
    :param additional_config_files: A list of paths to additional files to load as configuration.
        These will override any values found in the :file:`nunavut.lang.properties.yaml` file and files
        appearing later in this list will override value found in earlier entries.
    :type additional_config_files: typing.List[pathlib.Path]
    :param bool omit_serialization_support_for_target: If True then generators should not include
        serialization routines, types, or support libraries for the target language.
    :param typing.Optional[typing.Mapping[str, typing.Any]] language_options: Opaque arguments passed through to the
                target :class:`nunavut.lang.Language` object.
    :param bool include_experimental_languages: If True, expose languages with experimental (non-stable) support.
    :raises ValueError: If extension is None and no target language was provided.
    :raises KeyError: If the target language is not known.
    """

    def __init__(
        self,
        target_language: typing.Optional[str] = None,
        extension: typing.Optional[str] = None,
        namespace_output_stem: typing.Optional[str] = None,
        additional_config_files: typing.List[pathlib.Path] = [],
        omit_serialization_support_for_target: bool = True,
        language_options: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        include_experimental_languages: bool = True,
    ):
        self._extension = extension
        self._namespace_output_stem = namespace_output_stem
        self._ln_loader = LanguageLoader(*additional_config_files)
        self._config = self._ln_loader.config
        self._languages = dict()  # type: typing.Dict[str, Language]

        # create target language, if there is one.
        self._target_language = None
        if target_language is not None:
            try:
                self._target_language = self._ln_loader.load_language(
                    target_language, omit_serialization_support_for_target, language_options=language_options
                )
            except ImportError:
                raise KeyError("{} is not a supported language".format(target_language))
            if not (self._target_language.stable_support or include_experimental_languages):
                raise ValueError(
                    "{} support is only experimental, but experimental language support is not enabled".format(
                        target_language
                    )
                )
            if namespace_output_stem is None:
                self._namespace_output_stem = self._target_language.namespace_output_stem

            target_language_section_name = "nunavut.lang.{}".format(target_language)
            if self._namespace_output_stem is not None:
                self._config.set(target_language_section_name, "namespace_file_stem", self._namespace_output_stem)
            if extension is not None:
                self._config.set(target_language_section_name, "extension", extension)
            self._languages[target_language] = self._target_language

        # create remaining languages
        remaining_languages = set(self.get_supported_language_names()) - set((target_language,))
        self._populate_languages(remaining_languages, include_experimental_languages)

    def _populate_languages(self, language_names: typing.Iterable[str], include_experimental: bool) -> None:
        for language_name in language_names:
            try:
                lang = self._ln_loader.load_language(language_name, False)
                if lang.stable_support or include_experimental:
                    self._languages[language_name] = lang
            except ImportError:
                raise KeyError("{} is not a supported language".format(language_name))

    def get_language(self, key_or_module_name: str) -> Language:
        """
        Get a :class:`Language` object for a given language identifier.

        :param str key_or_module_name: Either one of the Nunavut mnemonics for a supported language or
            the ``__name__`` of one of the ``nunavut.lang.[language]`` python modules.
        :return: A :class:`Language` object cached by this context.
        :rtype: Language
        """
        if key_or_module_name is None or len(key_or_module_name) == 0:
            raise ValueError("key argument is required.")
        key = key_or_module_name[13:] if key_or_module_name.startswith("nunavut.lang.") else key_or_module_name
        return self.get_supported_languages()[key]

    def get_supported_language_names(self) -> typing.Iterable[str]:
        """Get a list of target languages supported by Nunavut.

        :return: An iterable of strings which are languages with special
            support within Nunavut templates.
        """
        return [s[13:] for s in self._config.sections() if s.startswith("nunavut.lang.")]

    def get_output_extension(self) -> str:
        """
        Gets the output extension to use regardless of a target language being available or not.

        :return: A file extension name with a leading dot.
        """
        if self._extension is not None:
            return self._extension
        elif self._target_language is not None:
            return self._target_language.extension
        else:
            raise RuntimeError(
                "No extension was provided and no target language was set. Cannot determine the extension to use."
            )

    def get_default_namespace_output_stem(self) -> typing.Optional[str]:
        """
        The filename stem to give to Namespace output files if emitted or None if there was none
        specified and there is no target language.

        :return: A file name stem or None
        """
        return self._namespace_output_stem

    def get_target_language(self) -> typing.Optional[Language]:
        """
        Returns the target language configured on this object or None
        if no target language was specified.
        """
        return self._target_language

    def filter_id_for_target(self, instance: typing.Any, id_type: str = "any") -> str:
        """
        A filter that will transform a given string or pydsdl identifier into a valid identifier in the target language.
        A default transformation is applied if no target language is set.

        :param any instance:        Any object or data that either has a name property or can be converted to a string.
        :param str id_type:         A type of identifier. This is different for each language.
                                    Use 'any' to apply stropping rules for all identifier types to the instance.
        :return: A token that is a valid identifier in the target language, is not a reserved keyword, and is
                 transformed in a deterministic manner based on the provided instance.
        """
        if self._target_language is not None:
            return self._target_language.filter_id(instance, id_type)
        else:
            return Language.default_filter_id_for_target(instance)

    def get_supported_languages(self) -> typing.Dict[str, Language]:
        """
        Returns a collection of available language support objects.
        """
        return self._languages

    @property
    def config(self) -> LanguageConfig:
        return self._config



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Abstractions around template engine internals.
"""
import functools
import inspect
import types
import typing

import nunavut.lang

ENVIRONMENT_FILTER_ATTRIBUTE_NAME = "environmentfilter"
"""
For now this is set to a value that is internally compatible with the
embedded version of jinja2 we use in Nunavut. If you use this variable
instead of its current value you will be insulated from this.
"""

CONTEXT_FILTER_ATTRIBUTE_NAME = "contextfilter"
"""
For now this is set to a value that is internally compatible with the
embedded version of jinja2 we use in Nunavut. If you use this variable
instead of its current value you will be insulated from this.
"""

LANGUAGE_FILTER_ATTRIBUTE_NAME = "nv_languagefilter"
"""
Nunavut-specific attribute for filters or tests that take their :class:`nunavut.lang.Language`
as the first argument.
"""


class SupportsTemplateContext:
    """
    Provided as a pseudo
    `protocol <https://mypy.readthedocs.io/en/latest/protocols.html#simple-user-defined-protocols/>`_.
    (in anticipation of that becoming part of the core Python typing someday).
    """


def template_environment_filter(filter_func: typing.Callable) -> typing.Callable[..., str]:
    """
    Decorator for marking environment dependent filters.
    An object supporting the :class:`SupportsTemplateEnv` protocol
    will be passed to the filter as the first argument.
    """
    setattr(filter_func, ENVIRONMENT_FILTER_ATTRIBUTE_NAME, True)
    return filter_func


def template_context_filter(filter_func: typing.Callable) -> typing.Callable[..., str]:
    """
    Decorator for marking context dependent filters.
    An object supporting the :class:`SupportsTemplateContext` protocol
    will be passed to the filter as the first argument.

    Note that any template that uses such a filter will make the jinja "frame"
    the filter appears within volatile and therefore unable to be optimized.
    """
    setattr(filter_func, CONTEXT_FILTER_ATTRIBUTE_NAME, True)
    return filter_func


def template_volatile_filter(filter_func: typing.Callable) -> typing.Callable[..., str]:
    """
    Decorator for marking a filter as volatile therefore disabling optimizations for the
    frame it appears within.
    An opaque object will be passed to the filter as the first argument.
    """
    setattr(filter_func, CONTEXT_FILTER_ATTRIBUTE_NAME, True)
    return filter_func


LanguageFilterReturnType = typing.TypeVar("LanguageFilterReturnType")


class GenericTemplateLanguageFilter(typing.Generic[LanguageFilterReturnType]):
    """
    Decorator for marking template filters that take a :class:`nunavut.lang.Language` object
    as the first argument with a generic return type.
    """

    def __init__(self, language_name_or_module: str):
        self._language_name_or_module = language_name_or_module

    def __call__(
        self, filter_func: typing.Callable[..., LanguageFilterReturnType]
    ) -> typing.Callable[..., LanguageFilterReturnType]:
        self._annotate_function(filter_func)
        return filter_func

    def _annotate_function(self, filter_func: typing.Callable[..., typing.Any]) -> None:
        setattr(filter_func, LANGUAGE_FILTER_ATTRIBUTE_NAME, self._language_name_or_module)


class template_language_filter(GenericTemplateLanguageFilter[str]):
    """
    Decorator for marking template filters that take a :class:`nunavut.lang.Language` object
    as the first argument.
    """

    pass


class template_language_list_filter(GenericTemplateLanguageFilter[typing.List[str]]):
    """
    Decorator for marking template filters that take a :class:`nunavut.lang.Language` object
    as the first argument and return a list of strings.
    """

    pass


class template_language_int_filter(GenericTemplateLanguageFilter[int]):
    """
    Decorator for marking template filters that take a :class:`nunavut.lang.Language` object
    as the first argument and return an integer.
    """

    pass


class template_language_test(GenericTemplateLanguageFilter[bool]):
    """
    Decorator for marking template tests that take a :class:`nunavut.lang.Language` object
    as the first argument.
    """

    pass


# +-------------------------------------------------------------------------------------------------------------------+
# | LanguageEnvironment
# +-------------------------------------------------------------------------------------------------------------------+


class LanguageEnvironment:
    """
    Data structure defining stuff contributed to a template environment for a given :class:`Language`.
    """

    TEST_NAME_PREFIX = "is_"
    FILTER_NAME_PREFIX = "filter_"
    USES_QUERY_PREFIX = "uses_"

    @classmethod
    def is_test_name(cls, callable_name: typing.Optional[str]) -> bool:
        return (
            callable_name is not None
            and len(callable_name) >= len(cls.TEST_NAME_PREFIX)
            and callable_name.startswith(cls.TEST_NAME_PREFIX)
        )

    @classmethod
    def is_filter_name(cls, callable_name: typing.Optional[str]) -> bool:
        return (
            callable_name is not None
            and len(callable_name) >= len(cls.FILTER_NAME_PREFIX)
            and callable_name.startswith(cls.FILTER_NAME_PREFIX)
        )

    @classmethod
    def is_uses_query_name(cls, callable_name: typing.Optional[str]) -> bool:
        return (
            callable_name is not None
            and len(callable_name) >= len(cls.USES_QUERY_PREFIX)
            and callable_name.startswith(cls.USES_QUERY_PREFIX)
        )

    LanguageListT = typing.TypeVar(
        "LanguageListT", typing.AbstractSet[nunavut.lang.Language], typing.ValuesView[nunavut.lang.Language]
    )

    def __init__(self, language_name: str) -> None:
        self._language_name = language_name
        self._tests = dict()  # type: typing.Dict[str, typing.Callable]
        self._filters = dict()  # type: typing.Dict[str, typing.Callable]
        self._uses_queries = dict()  # type: typing.Dict[str, typing.Callable]

    @property
    def language_name(self) -> str:
        return self._language_name

    @property
    def tests(self) -> typing.Mapping[str, typing.Callable]:
        return self._tests

    @property
    def filters(self) -> typing.Mapping[str, typing.Callable]:
        return self._filters

    @property
    def uses_queries(self) -> typing.Mapping[str, typing.Callable]:
        return self._uses_queries

    @classmethod
    def _parse_callable_name(
        cls, callable: typing.Callable, callable_name: typing.Optional[str] = None
    ) -> typing.Tuple[typing.Optional[str], str]:
        if callable_name is None:
            if isinstance(callable, functools.partial):
                callable_name = callable.func.__name__
            else:
                callable_name = callable.__name__

        if cls.is_test_name(callable_name):
            prefix = cls.TEST_NAME_PREFIX  # type: typing.Optional[str]
            method_name = callable_name[len(cls.TEST_NAME_PREFIX) :]
        elif cls.is_filter_name(callable_name):
            prefix = cls.FILTER_NAME_PREFIX
            method_name = callable_name[len(cls.FILTER_NAME_PREFIX) :]
        elif cls.is_uses_query_name(callable_name):
            prefix = cls.USES_QUERY_PREFIX
            method_name = callable_name[len(cls.USES_QUERY_PREFIX) :]
        else:
            prefix = None
            method_name = callable_name

        return (prefix, method_name)

    @classmethod
    def handle_conventional_methods(
        cls,
        callable: typing.Callable,
        callable_name: typing.Optional[str] = None,
        supported_languages: typing.Optional[LanguageListT] = None,
    ) -> typing.Tuple[typing.Optional[str], str, typing.Callable]:
        """
        Processes method objects that utilize the nunavut convention of ``is_``, ``filter_``, or ``uses_`` prefixes.
        Also wraps the method in a partial if it requested the language as the first argument.

        :param str callable_name: If provided this is the name used to process the callable otherwise
                                the ``__name__`` property is used from the callable itself.
        :return: A 3-tuple with the prefix, method name without prefix, and the method or partial. If the first
                                element is ``None`` then the callable was not a conventional method.
        """

        prefix, method_name = cls._parse_callable_name(callable, callable_name)

        resolved_callable = None  # type: typing.Optional[typing.Callable]
        if hasattr(callable, LANGUAGE_FILTER_ATTRIBUTE_NAME):
            callable_language_name = getattr(callable, LANGUAGE_FILTER_ATTRIBUTE_NAME)

            if supported_languages is not None:
                for language in supported_languages:
                    if language.get_templates_package_name() == callable_language_name:
                        resolved_callable = functools.partial(callable, language)
                        break
            if resolved_callable is None:
                raise RuntimeWarning(
                    'Language callable "{}", required an unsupported language({})'.format(
                        method_name, callable_language_name
                    )
                )
        else:
            resolved_callable = callable

        return (prefix, method_name, resolved_callable)

    # +---------------------------------------------------------------------------------------------------------------+
    # | DATA MODEL
    # +---------------------------------------------------------------------------------------------------------------+
    def __getitem__(self, key: str) -> typing.Dict[str, typing.Callable]:
        key_case_insensitive = key.lower()
        if key_case_insensitive.startswith("is"):
            return self._tests
        elif key_case_insensitive.startswith("filter"):
            return self._filters
        elif key_case_insensitive.startswith("uses"):
            return self._uses_queries
        else:
            raise KeyError("Key {} was not supported by this object.".format(key))

    # +---------------------------------------------------------------------------------------------------------------+
    # | FACTORY
    # +---------------------------------------------------------------------------------------------------------------+
    @classmethod
    def find_all_conventional_methods_in_language_module(
        cls, language: nunavut.lang.Language, all_languages: LanguageListT, language_module: "types.ModuleType"
    ) -> "LanguageEnvironment":
        results = LanguageEnvironment(language.name)

        callables = inspect.getmembers(language_module, inspect.isfunction)
        for function_tuple in callables:
            result = cls.handle_conventional_methods(function_tuple[1], supported_languages=all_languages)
            if result[0] is not None:
                results[result[0]][result[1]] = result[2]

        return results



#
# Copyright (C) 2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Contains the Jinja templates to generate HTML documentation.
"""

__version__ = "1.0.0"
"""Version of the HTML docgen template set."""
